
import urllib.request
import datetime
import json
import re
import pandas as pd
from flask import Flask , request, render_template, Markup, send_file
from openpyxl import Workbook
from xhtml2pdf import pisa

app = Flask(__name__)

def downloadSummary(query_text):
    """
    FHIRサーバから退院時サマリーのデータをダウンロードする

    Parameters
    ----------
    query_text: string
        FHIRサーバを検索するためのクエリ文字列

    Returns
    -------
    jsonobj: dict
        退院時サマリーのJSONデータを格納したディクショナリ
    """
    resturl = f'https://fhirtestjp.med.gunma-u.ac.jp/Bundle?{query_text}'
    jsonobj=None
    with urllib.request.urlopen(resturl) as response:
        fhir = response.read()
        jsonobj = json.loads(fhir.decode('utf-8'))
        #jsonstr = json.dumps(jsonobj, indent=2, ensure_ascii=False)

    # json_path = './summary_example.json'
    # with open(json_path, mode='r',encoding='utf-8') as json_file:
    #     jsonstr=json_file.read()
    #     jsonobj = json.loads(jsonstr)

    return jsonobj

def WriteSummaryExcel(jsondoc):
    """
    退院時サマリーのFHIRデータの内容をExcelファイルに書き込む

    Parameters
    ----------
    jsondoc: dict
        退院時サマリーのFHIRデータを格納したディクショナリ

    Returns
    -------
    excel_path: string
        退院時サマリーのFHIRデータを書き込んだExcelファイル名(xlsx形式)
    """

    xlsx_wb = Workbook()
    xlsx_ws = xlsx_wb['Sheet']
    xlsx_ws = xlsx_wb.active

    timsstampstr=datetime.datetime.now().strftime('%Y%m%d%H%M%S')
    excel_path=f'summary_{timsstampstr}.xlsx'

    lst=[p for p in jsondoc['entry'] if p['resource']['resourceType']=='Composition']
    subjectid=''
    encounterid=''
    autheridlst=[]
    if len(lst)>0 :
        composi=lst[0]['resource']

        xlsx_ws['A1'] = '文書ID'
        print(composi['identifier']['value'])
        xlsx_ws['B1'] = composi['identifier']['value']

        xlsx_ws['A2'] = '文書種類コード'
        print(composi['type']['coding'][0]['code'])
        xlsx_ws['B2'] = composi['type']['coding'][0]['code']

        xlsx_ws['A3'] = '文書種類名'
        print(composi['type']['coding'][0]['display'])
        xlsx_ws['B3'] = composi['type']['coding'][0]['display']

        xlsx_ws['A4'] = '文書タイトル'
        print(composi['title'])
        xlsx_ws['B4'] = composi['title']

        xlsx_ws['A5'] = '文書作成日'
        print(composi['date'])
        xlsx_ws['B5'] = composi['date']

        xlsx_ws['A6'] = '文書状態'
        print(composi['status'])
        xlsx_ws['B6'] = composi['status']

        subjectid=composi['subject']['reference']
        encounterid=composi['encounter']['reference']
        autheridlst=[i['reference'] for i in composi['author']]
    else:
        raise NameError('entryにCompositionなし')

    lst=[p for p in jsondoc['entry'] if p.get('fullUrl')==subjectid]
    if len(lst)>0 :
        pat=lst[0]['resource']

        xlsx_ws['A8'] = '患者ID'
        print(pat['identifier'][0]['value'])
        xlsx_ws['B8'] = pat['identifier'][0]['value']

        xlsx_ws['A9'] = '患者氏名'
        print(pat['name'][0]['text'])
        xlsx_ws['B9'] = pat['name'][0]['text']

        xlsx_ws['A10'] = '患者カナ'
        print(pat['name'][1]['text'])
        xlsx_ws['B10'] = pat['name'][1]['text']

        xlsx_ws['A11'] = '性別'
        print(pat['gender'])
        xlsx_ws['B11'] = pat['gender']

        xlsx_ws['A12'] = '生年月日'
        print(pat['birthDate'])
        xlsx_ws['B12'] = pat['birthDate']

        xlsx_ws['A13'] = '郵便番号'
        print(pat['address'][0]['postalCode'])
        xlsx_ws['B13'] = pat['address'][0]['postalCode']

        xlsx_ws['A14'] = '住所'
        print(pat['address'][0]['text'])
        xlsx_ws['B14'] = pat['address'][0]['text']
    else:
        raise NameError('entryにPatientなし')

    lst=[e for e in jsondoc['entry'] if e.get('fullUrl')==encounterid]
    if len(lst)>0 :
        enc=lst[0]['resource']
        
        xlsx_ws['A16'] = '入院日'
        print(enc['period']['start'])
        xlsx_ws['B16'] = enc['period']['start']
        
        xlsx_ws['A17'] = '退院日'
        print(enc['period']['end'])
        xlsx_ws['B17'] = enc['period']['end']

        xlsx_ws['A18'] = '入院理由'
        print(','.join(enc['reasonCode']))
        xlsx_ws['B18'] = ','.join(enc['reasonCode'])

        diaglst=sorted(enc['diagnosis'],key=lambda x: x['rank'])
        adlst=[]
        for diag in diaglst:
            adcd=[i for i in diag['use']['coding'] if i['code']=='AD']
            if(len(adcd)>0):
                adlst.append(diag['condition'])
        for condiid in adlst:
            condilst=[e for e in jsondoc['entry'] if e.get('fullUrl')==condiid]
            if len(condiid)>0:
                condi=condilst[0]['resource']
                xlsx_ws['A19'] = '入院時診断名'
                print(condi['code']['text'])
                xlsx_ws['B19'] = condi['code']['text']

                xlsx_ws['A20'] = '入院時診断日'
                print(condi['onsetDateTime'])
                xlsx_ws['B20'] = condi['onsetDateTime']
        ddlst=[]
        for diag in diaglst:
            ddcd=[i for i in diag['use']['coding'] if i['code']=='DD']
            if(len(ddcd)>0):
                ddlst.append(diag['condition'])
        for condiid in ddlst:
            condilst=[e for e in jsondoc['entry'] if e.get('fullUrl')==condiid]
            if len(condiid)>0:
                condi=condilst[0]['resource']
                xlsx_ws['A21'] = '退院時診断名'
                print(condi['code']['text'])
                xlsx_ws['B21'] = condi['code']['text']

                xlsx_ws['A22'] = '退院時診断日'
                print(condi['onsetDateTime'])
                xlsx_ws['B22'] = condi['onsetDateTime']
    else:
        raise NameError('entryにEncounterなし')

    for autherid in autheridlst:
        lst=[i for i in jsondoc['entry'] if i.get('fullUrl')==autherid and i.get('resource').get('resourceType')=='Organization']
        if len(lst)>0:
            orga=lst[0]['resource']

            xlsx_ws['A24'] = '病院コード'
            print(orga['identifier'][0]['value'])
            xlsx_ws['B24'] = orga['identifier'][0]['value']

            xlsx_ws['A25'] = '病院名'
            print(orga['name'])
            xlsx_ws['B25'] = orga['name']

        lst=[i for i in jsondoc['entry'] if i.get('fullUrl')==autherid and i.get('resource').get('resourceType')=='Practitioner']
        if len(lst)>0:
            pract=lst[0]['resource']

            xlsx_ws['A26'] = '医師ID'
            print(pract['identifier'][0]['value'])
            xlsx_ws['B26'] = pract['identifier'][0]['value']

            xlsx_ws['A27'] = '医師氏名'
            print(pract['name'][0]['text'])
            xlsx_ws['B27'] = pract['name'][0]['text']

            xlsx_ws['A28'] = '医師カナ'
            print(pract['name'][1]['text'])
            xlsx_ws['B28'] = pract['name'][1]['text']

    xlsx_wb.save(excel_path)

    return excel_path

def convertHtmlString(excel_file):
    """
    退院時サマリーのExcelファイルからHTMLの表データに変換する

    Parameters
    ----------
    excel_file: string
        退院時サマリーのFHIRデータが保存されているExcelファイル名(xlsx形式)

    Returns
    -------
    htmlstr: string
        退院時サマリーのFHIRデータをHTML表データの文字列
    """

    excel_df = pd.read_excel(excel_file, engine='openpyxl',header=None)

    htmllines=[]

    htmllines.append('<table border="1">\n')
    for i in range(len(excel_df)):
        htmllines.append("\t<tr>\n")
        for k in range(len(excel_df.columns)):
            val = excel_df.iloc[i, k]
            rowspan = 1
            if isinstance(val, str):
                val = re.sub("(\r\n)|(\n)", "<br />", val)
                l = 1
                while i+l < len(excel_df):
                    _val = excel_df.iloc[i+l, k]
                    if isinstance(_val, str):
                        break
                    l += 1
                rowspan = l
                htmllines.append(f'\t\t<td colspan="1" rowspan="{rowspan}">{val}</td>\n')
        htmllines.append("\t</tr>\n")
    htmllines.append('</table>\n')

    htmlstr=''.join(htmllines)
    return htmlstr

def writeSummaryPDF(excel_file):
    """
    退院時サマリーのExcelファイルをもとにPDFファイルに書き込む

    Parameters
    ----------
    excel_file: string
        退院時サマリーのFHIRデータが保存されているExcelファイル名(xlsx形式)

    Returns
    -------
    pdf_file: string
        退院時サマリーのFHIRデータを書き込んだPDFファイル名
    """

    htmllines=[]

    htmllines.append('<html><head>\n')
    htmllines.append('<meta http-equiv="content-type" content="text/html; charset=utf-8">\n')
    htmllines.append('<title>summary</title>\n')
    htmllines.append('<style> body { font-family: HeiseiKakuGo-W5 } </style>\n')
    htmllines.append('</head><body>\n')
    htmllines.append(convertHtmlString(excel_file))
    htmllines.append('\n')
    htmllines.append('</body></html>\n')

    htmlstr=''.join(htmllines)

    timsstampstr=datetime.datetime.now().strftime('%Y%m%d%H%M%S')
    pdf_file = f'summary_{timsstampstr}.pdf'
    with open(pdf_file, "w+b") as pdf:
            pisa_status = pisa.CreatePDF(src=htmlstr, dest=pdf)

    return pdf_file


@app.route('/')
def index():
    '''
    退院時サマリーの検索画面を表示する
    '''
    return render_template('index.html')

@app.route('/SummaryViewer', methods=["POST"])
def viewSummary():
    '''
    指定したパラメータの条件でFHIRサーバの退院時サマリーを検索し画面に表示する
    '''
    qt=request.form["querytext"] 

    jsondoc=downloadSummary(qt)
    excel_file=WriteSummaryExcel(jsondoc)

    sum_tbl_html=Markup(convertHtmlString(excel_file))
    return render_template('summary.html',query_text=qt,sum_tbl_html=sum_tbl_html)

@app.route('/SummaryExcel', methods=["GET"])
def downloadSummaryExcel():
    '''
    指定したパラメータの条件でFHIRサーバの退院時サマリーを検索しExcelファイルにしてダウンロードする
    '''
    qt=request.query_string

    jsondoc=downloadSummary(qt)
    excel_file=WriteSummaryExcel(jsondoc)

    return send_file(excel_file, as_attachment=True,
                        attachment_filename=excel_file,
                        mimetype='application/vnd.openxmlformats-officedocument.spreadsheetml.sheet')

@app.route('/SummaryPDF', methods=["GET"])
def downloadSummaryPDF():
    '''
    指定したパラメータの条件でFHIRサーバの退院時サマリーを検索しPDFファイルにしてダウンロードする
    '''
    qt=request.query_string

    jsondoc=downloadSummary(qt)
    excel_file=WriteSummaryExcel(jsondoc)
    pdf_file=writeSummaryPDF(excel_file)

    return send_file(pdf_file, as_attachment=True,
                        attachment_filename=pdf_file,
                        mimetype='application/pdf')

if __name__ == "__main__":
    app.run(debug=True)

